// Flutter Packages
import 'package:flutter/material.dart';

// Dart Packages
import 'dart:async';

// Firebase Packages
import 'package:firebase_database/firebase_database.dart';

// Data Models
import 'models/view_type.dart';
import 'models/wrap_options.dart';

// Widgets
import 'widgets/defaults/bottom_loader.dart';
import 'widgets/defaults/empty_screen.dart';
import 'widgets/defaults/initial_loader.dart';
import 'widgets/views/build_pagination.dart';

/// A [StreamBuilder] that automatically loads more data when the user scrolls
/// to the bottom.
///
/// Optimized for [FirebaseDatabase] with fields like `createdAt` and
/// `timestamp` to sort the data.
///
/// Supports realtime updates and updates to loaded data.
///
/// Data can be represented in a [ListView], [GridView] or scollable [Wrap].
class RealtimeDBPagination extends StatefulWidget {
  /// Creates a [StreamBuilder] widget that automatically loads more data when
  /// the user scrolls to the bottom.
  ///
  /// Optimized for [FirebaseDatabase] with fields like `createdAt` and
  /// `timestamp` to sort the data.
  ///
  /// Supports realtime updates and updates to loaded data.
  ///
  /// Data can be represented in a [ListView], [GridView] or scollable [Wrap].
  const RealtimeDBPagination({
    Key? key,
    required this.query,
    required this.itemBuilder,
    this.limit = 10,
    this.viewType = ViewType.list,
    this.isRealtime = true,
    this.gridDelegate = const SliverGridDelegateWithFixedCrossAxisCount(
      crossAxisCount: 2,
      childAspectRatio: 1,
    ),
    this.wrapOptions = const WrapOptions(),
    this.onEmpty = const EmptyScreen(),
    this.bottomLoader = const BottomLoader(),
    this.initialLoader = const InitialLoader(),
  }) : super(key: key);

  /// The query to use to fetch data from Firebase Realtime Database.
  final Query query;

  /// The builder to use to build the items in the list.
  ///
  /// The builder is passed the build context, snapshot of data and index of
  /// the item in the list.
  final Widget Function(BuildContext, DataSnapshot, int) itemBuilder;

  /// The number of items to fetch from Firebase Realtime Database at once.
  ///
  /// Defaults to `10`.
  final int limit;

  /// The type of view to use for the list.
  ///
  /// Defaults to [ViewType.list].
  final ViewType viewType;

  /// Whether to fetch newly added items as they are added to
  /// Firebase Realtime Database.
  ///
  /// Defaults to `true`.
  final bool isRealtime;

  /// The delegate to use for the [GridView].
  ///
  /// Defaults to [SliverGridDelegateWithFixedCrossAxisCount].
  final SliverGridDelegate gridDelegate;

  /// The [Wrap] widget properties to use.
  ///
  /// Defaults to [WrapOptions].
  final WrapOptions wrapOptions;

  /// The widget to use when data is empty.
  ///
  /// Defaults to [EmptyScreen].
  final Widget onEmpty;

  /// The widget to use when more data is loading.
  ///
  /// Defaults to [BottomLoader].
  final Widget bottomLoader;

  /// The widget to use when data is loading initially.
  ///
  /// Defaults to [InitialLoader].
  final Widget initialLoader;

  @override
  State<RealtimeDBPagination> createState() => _RealtimeDBPaginationState();
}

/// The state of the [RealtimeDBPagination] widget.
class _RealtimeDBPaginationState extends State<RealtimeDBPagination> {
  /// All the data that has been loaded from Firebase Realtime Database.
  final List<DataSnapshot> _data = [];

  /// Snapshot subscription for the query.
  ///
  /// Also handles updates to loaded data.
  StreamSubscription<DatabaseEvent>? _streamSub;

  /// Snapshot subscription for the query to handle newly added data.
  StreamSubscription<DatabaseEvent>? _realtimeStreamSub;

  /// [ScrollController] to listen to scroll end and load more data.
  final ScrollController _controller = ScrollController();

  /// Whether initial data is loading.
  bool _isInitialLoading = true;

  /// Whether more data is loading.
  bool _isFetching = false;

  /// Whether the end for given query has been reached.
  ///
  /// This is used to determine if more data should be loaded when the user
  /// scrolls to the bottom.
  bool _isEnded = false;

  /// Loads more data from Firebase Realtime Database and handles
  /// updates to loaded data.
  ///
  /// Setting [getMore] to `false` will only set listener for the
  /// currently loaded data.
  Future<void> _loadData({bool getMore = true}) async {
    // To cancel previous updates listener when new one is set.
    final tempSub = _streamSub;

    if (getMore) setState(() => _isFetching = true);

    int docsLimit = _data.length + (getMore ? widget.limit : 0);
    Query docsQuery = widget.query.limitToFirst(docsLimit);
    if (_data.isNotEmpty) {
      docsQuery = docsQuery.startAt(_data.first);
    }

    _streamSub = docsQuery.onValue.listen((DatabaseEvent snapshot) async {
      await tempSub?.cancel();

      _data.clear();
      _data.addAll(snapshot.snapshot.children);

      // To set new updates listener for the existing data
      // or to set new realtime listener if the first data node is removed.
      bool isDataRemoved = snapshot.type == DatabaseEventType.childRemoved;

      _isFetching = false;
      if (!_isEnded && !isDataRemoved) {
        _isEnded = snapshot.snapshot.children.length < docsLimit;
      }

      if (isDataRemoved || _isInitialLoading) {
        _isInitialLoading = false;
        if (snapshot.snapshot.children.isNotEmpty) {
          // Set updates listener for the existing data starting from the
          // first data node only.
          await _loadData(getMore: false);
        } else {
          _streamSub?.cancel();
        }
        if (widget.isRealtime) _setRealtimeListner();
      }

      if (mounted) setState(() {});
    });
  }

  /// Sets the realtime listener for the query.
  ///
  /// Fires when new data is added to the query.
  Future<void> _setRealtimeListner() async {
    // To cancel previous realtime listener when new one is set.
    final tempSub = _realtimeStreamSub;

    Query latestDocQuery = widget.query.limitToFirst(1);
    if (_data.isNotEmpty) {
      latestDocQuery = latestDocQuery.endBefore(_data.first);
    }

    _realtimeStreamSub = latestDocQuery.onValue.listen(
      (DatabaseEvent snapshot) async {
        await tempSub?.cancel();
        if (snapshot.snapshot.children.isEmpty) return;

        _data.insert(0, snapshot.snapshot.children.first);

        // To handle newly added data after this curently loaded data.
        await _setRealtimeListner();

        // Set updates listener for the newly added data.
        _loadData(getMore: false);
      },
    );
  }

  /// To handle scroll end event and load more data.
  void _scrollListener() {
    final ScrollPosition position = _controller.position;
    if (position.pixels >= (position.maxScrollExtent - 50) &&
        !_isInitialLoading &&
        !_isFetching &&
        !_isEnded) {
      _loadData();
    }
  }

  @override
  void initState() {
    super.initState();
    _loadData();
    _controller.addListener(_scrollListener);
  }

  @override
  void dispose() {
    _streamSub?.cancel();
    _realtimeStreamSub?.cancel();
    _controller.removeListener(_scrollListener);
    _controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _isInitialLoading
        ? widget.initialLoader
        : _data.isEmpty
            ? widget.onEmpty
            : BuildPagination(
                controller: _controller,
                items: _data,
                itemBuilder: widget.itemBuilder,
                isLoading: _isFetching,
                bottomLoader: widget.bottomLoader,
                gridDelegate: widget.gridDelegate,
                viewType: widget.viewType,
                wrapOptions: widget.wrapOptions,
              );
  }
}
