// Flutter Packages
import 'package:flutter/material.dart';

// Data Models
import '../../models/view_type.dart';
import '../../models/wrap_options.dart';

// Widgets
import '../defaults/bottom_loader.dart';

/// A [ScrollView] to use for the provided [items].
///
/// The [items] are loaded into the [ScrollView] based on the [viewType].
class BuildPagination<T> extends StatelessWidget {
  /// Creates a [ScrollView] to use for the provided [items].
  ///
  /// The [items] are rendered in the [ScrollView] using the [itemBuilder].
  ///
  /// The [viewType] determines the type of [ScrollView] to use.
  const BuildPagination({
    Key? key,
    required this.controller,
    required this.items,
    required this.itemBuilder,
    required this.isLoading,
    this.gridDelegate = const SliverGridDelegateWithFixedCrossAxisCount(
      crossAxisCount: 2,
      childAspectRatio: 1,
    ),
    this.viewType = ViewType.list,
    this.bottomLoader = const BottomLoader(),
    this.wrapOptions = const WrapOptions(),
  }) : super(key: key);

  /// The type of [ScrollView] to use.
  final ViewType viewType;

  /// The scroll controller to handle the scroll events.
  final ScrollController controller;

  /// The items to display in the [ScrollView].
  final List<T> items;

  /// The builder to use to render the items.
  final Widget Function(BuildContext, T, int) itemBuilder;

  /// Whether more [items] are being loaded.
  final bool isLoading;

  /// The delegate to use for the [GridView].
  final SliverGridDelegate gridDelegate;

  /// A [Widget] to show when more [items] are being loaded.
  final Widget bottomLoader;

  /// The options to use for the [Wrap].
  final WrapOptions wrapOptions;

  @override
  Widget build(BuildContext context) {
    switch (viewType) {
      case ViewType.list:
        return ListView.builder(
          controller: controller,
          itemCount: items.length + (isLoading ? 1 : 0),
          itemBuilder: (BuildContext context, int index) {
            if (index >= items.length) return bottomLoader;

            return itemBuilder(context, items[index], index);
          },
        );

      case ViewType.grid:
        return GridView.builder(
          controller: controller,
          itemCount: items.length + (isLoading ? 1 : 0),
          itemBuilder: (BuildContext context, int index) {
            if (index >= items.length) return bottomLoader;

            return itemBuilder(context, items[index], index);
          },
          gridDelegate: gridDelegate,
        );

      case ViewType.wrap:
        return SingleChildScrollView(
          controller: controller,
          child: Wrap(
            direction: wrapOptions.direction,
            alignment: wrapOptions.alignment,
            spacing: wrapOptions.spacing,
            runAlignment: wrapOptions.runAlignment,
            runSpacing: wrapOptions.runSpacing,
            crossAxisAlignment: wrapOptions.crossAxisAlignment,
            textDirection: wrapOptions.textDirection,
            verticalDirection: wrapOptions.verticalDirection,
            clipBehavior: wrapOptions.clipBehavior,
            children: List.generate(
              items.length + (isLoading ? 1 : 0),
              (int index) {
                if (index >= items.length) return bottomLoader;

                return itemBuilder(context, items[index], index);
              },
            ),
          ),
        );
    }
  }
}
